/*
*   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*       http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*/

#include <cmath>
#include <limits>

#include "numerics.h"
#include "utils.h"

using namespace std;

static double relerr(double expected, double actual) {
    return fabs((expected - actual)/actual);
}

static void test_draw(void) {
    using numerics::draw_sample_unnormalized;
    const double epsilon = std::numeric_limits<double>::epsilon();
    const double denorm_min = std::numeric_limits<double>::denorm_min();

    vector<double> weights;

    weights.resize(4);
    weights[0] = log(1);
    weights[1] = log(2);
    weights[2] = log(4);
    weights[3] = log(8);
    assert(draw_sample_unnormalized(weights, 0.0) == 0);
    assert(draw_sample_unnormalized(weights, 0.1) == 1);
    assert(draw_sample_unnormalized(weights, 0.2) == 2);
    assert(draw_sample_unnormalized(weights, 0.3) == 2);
    assert(draw_sample_unnormalized(weights, 0.4) == 2);
    assert(draw_sample_unnormalized(weights, 0.5) == 3);
    assert(draw_sample_unnormalized(weights, 0.6) == 3);
    assert(draw_sample_unnormalized(weights, 0.7) == 3);
    assert(draw_sample_unnormalized(weights, 0.8) == 3);
    assert(draw_sample_unnormalized(weights, 0.9) == 3);
    assert(draw_sample_unnormalized(weights, 1 - epsilon/2) == 3);

    weights.resize(9);
    weights[0] = -denorm_min;
    for (size_t i = 1; i < weights.size(); i++)
	weights[i] = -1;
    assert(draw_sample_unnormalized(weights, 0) == 0);
    assert(0 <= draw_sample_unnormalized(weights, 1 - epsilon/2));
    assert((size_t)draw_sample_unnormalized(weights, 1 - epsilon/2) <
      weights.size());
}

static void test_linspace(void) {
    vector<double> v;

    v = linspace(42, 43, 2);
    assert(v.size() == 2);
    assert(v[0] == 42);
    assert(v[1] == 43);

    v = linspace(42, 42, 2);
    assert(v.size() == 2);
    assert(v[0] == 42);
    assert(v[1] == 42);

    v = linspace(42, 43, 3);
    assert(v.size() == 3);
    assert(v[0] == 42);
    assert(v[1] == 42.5);
    assert(v[2] == 43);

    v = linspace(42, 42, 3);
    assert(v.size() == 3);
    assert(v[0] == 42);
    assert(v[1] == 42);
    assert(v[2] == 42);

    v = linspace(0, 1, 7);
    assert(v.size() == 7);
    assert(v[6] == 1);
}

static void test_log_linspace(void) {
    vector<double> v;

    v = log_linspace(1, 8, 4);
    assert(v[0] == 1);
    assert(v[1] == 2);
    assert(v[2] == 4);
    assert(v[3] == 8);

    v = log_linspace(0, 42, 4);
    assert(v[0] == std::numeric_limits<double>::min());
    assert(v[3] == 42);

    v = log_linspace(0, 0, 42);
    for (size_t i = 0; i < v.size(); i++)
        assert(v[i] == std::numeric_limits<double>::min());
}

static void test_logaddexp(void) {
    vector<double> v(1);

    v[0] = 1000;
    assert(numerics::logaddexp(v) == 1000);
    v[0] = -1000;
    assert(numerics::logaddexp(v) == -1000);
}

// The values of the modified Bessel function of the first kind with
// nu = 0, 1 tend to overflow outside [-709, +709], depending on the
// exact method used to evaluate it.  So just test a uniform grid of
// values there, and a normal grid around 0 and 15 where logic in the
// implementation changes, generated by the program bessel.cpp using
// boost 1.54.0.1 on Ubuntu 14.04 amd64.

#define arraycount(A) (sizeof(A)/sizeof(*(A)))

static void test_bessel(void) {
    static const double i0e[][2] = {
        // Uniform [-709, +709] grid
        { -7.09000000000000000e+02, 1.23154770670165400e+306 },
        { -6.94820000000000050e+02, 8.64060224208248464e+299 },
        { -6.80639999999999986e+02, 6.06355486411944736e+293 },
        { -6.66460000000000036e+02, 4.25603261442151850e+287 },
        { -6.52279999999999973e+02, 2.98800266537820785e+281 },
        { -6.38100000000000023e+02, 2.09826211076560832e+275 },
        { -6.23919999999999959e+02, 1.47382473286544928e+269 },
        { -6.09740000000000009e+02, 1.03548598244646494e+263 },
        { -5.95559999999999945e+02, 7.27713072718670134e+256 },
        { -5.81379999999999995e+02, 5.11563284594497492e+250 },
        { -5.67200000000000045e+02, 3.59722733532623695e+244 },
        { -5.53019999999999982e+02, 2.53030155402530434e+238 },
        { -5.38840000000000032e+02, 1.78040834846892484e+232 },
        { -5.24659999999999968e+02, 1.25319176676307875e+226 },
        { -5.10480000000000018e+02, 8.82417822255562042e+219 },
        { -4.96300000000000011e+02, 6.21582516359315088e+213 },
        { -4.82120000000000005e+02, 4.38026942558973507e+207 },
        { -4.67939999999999998e+02, 3.08809732255943058e+201 },
        { -4.53759999999999991e+02, 2.17811516863629755e+195 },
        { -4.39579999999999984e+02, 1.53703266180069692e+189 },
        { -4.25399999999999977e+02, 1.08520475645193952e+183 },
        { -4.11220000000000027e+02, 7.66623229421532392e+176 },
        { -3.97040000000000020e+02, 5.41889743468751645e+170 },
        { -3.82860000000000014e+02, 3.83281085940988227e+164 },
        { -3.68680000000000007e+02, 2.71282832415018515e+158 },
        { -3.54500000000000000e+02, 1.92153869133222393e+152 },
        { -3.40319999999999993e+02, 1.36214757916307467e+146 },
        { -3.26139999999999986e+02, 9.66444833663966689e+139 },
        { -3.11960000000000036e+02, 6.86343433320022117e+133 },
        { -2.97780000000000030e+02, 4.87927994849656570e+127 },
        { -2.83600000000000023e+02, 3.47267211223380114e+121 },
        { -2.69420000000000016e+02, 2.47466453728772990e+115 },
        { -2.55240000000000009e+02, 1.76592545379549455e+109 },
        { -2.41060000000000002e+02, 1.26212092436926333e+103 },
        { -2.26879999999999995e+02, 9.03615804951987665e+96 },
        { -2.12699999999999989e+02, 6.48214067380827254e+90 },
        { -1.98519999999999982e+02, 4.66039395231502459e+84 },
        { -1.84340000000000032e+02, 3.35923482885829892e+78 },
        { -1.70159999999999968e+02, 2.42856843027191752e+72 },
        { -1.55980000000000018e+02, 1.76188701663370574e+66 },
        { -1.41799999999999955e+02, 1.28355264186649118e+60 },
        { -1.27620000000000005e+02, 9.39808860168245738e+53 },
        { -1.13440000000000055e+02, 6.92426473348450486e+47 },
        { -9.92599999999999909e+01, 5.14212824438065119e+41 },
        { -8.50800000000000409e+01, 3.85844851766619089e+35 },
        { -7.08999999999999773e+01, 2.93654646727135523e+29 },
        { -5.67200000000000273e+01, 2.28134342069862139e+23 },
        { -4.25399999999999636e+01, 1.83100740013884288e+17 },
        { -2.83600000000000136e+01, 1.55990347062181732e+11 },
        { -1.41800000000000637e+01, 1.53936747894937813e+05 },
        { 0.00000000000000000e+00, 1.00000000000000000e+00 },
        { 1.41799999999999500e+01, 1.53936747894920962e+05 },
        { 2.83600000000000136e+01, 1.55990347062181732e+11 },
        { 4.25399999999999636e+01, 1.83100740013884288e+17 },
        { 5.67200000000000273e+01, 2.28134342069862139e+23 },
        { 7.08999999999999773e+01, 2.93654646727135523e+29 },
        { 8.50799999999999272e+01, 3.85844851766575481e+35 },
        { 9.92599999999999909e+01, 5.14212824438065119e+41 },
        { 1.13439999999999941e+02, 6.92426473348372115e+47 },
        { 1.27620000000000005e+02, 9.39808860168245738e+53 },
        { 1.41799999999999955e+02, 1.28355264186649118e+60 },
        { 1.55980000000000018e+02, 1.76188701663370574e+66 },
        { 1.70159999999999968e+02, 2.42856843027191752e+72 },
        { 1.84340000000000032e+02, 3.35923482885829892e+78 },
        { 1.98519999999999982e+02, 4.66039395231502459e+84 },
        { 2.12699999999999932e+02, 6.48214067380790436e+90 },
        { 2.26879999999999995e+02, 9.03615804951987665e+96 },
        { 2.41059999999999945e+02, 1.26212092436919171e+103 },
        { 2.55240000000000009e+02, 1.76592545379549455e+109 },
        { 2.69419999999999959e+02, 2.47466453728758948e+115 },
        { 2.83600000000000023e+02, 3.47267211223380114e+121 },
        { 2.97779999999999973e+02, 4.87927994849628866e+127 },
        { 3.11960000000000036e+02, 6.86343433320022117e+133 },
        { 3.26139999999999873e+02, 9.66444833663857112e+139 },
        { 3.40319999999999936e+02, 1.36214757916299726e+146 },
        { 3.54500000000000000e+02, 1.92153869133222393e+152 },
        { 3.68680000000000064e+02, 2.71282832415033880e+158 },
        { 3.82859999999999900e+02, 3.83281085940944752e+164 },
        { 3.97039999999999964e+02, 5.41889743468720968e+170 },
        { 4.11220000000000027e+02, 7.66623229421532392e+176 },
        { 4.25400000000000091e+02, 1.08520475645206288e+183 },
        { 4.39579999999999927e+02, 1.53703266180060961e+189 },
        { 4.53759999999999991e+02, 2.17811516863629755e+195 },
        { 4.67940000000000055e+02, 3.08809732255960626e+201 },
        { 4.82119999999999891e+02, 4.38026942558923720e+207 },
        { 4.96299999999999955e+02, 6.21582516359279686e+213 },
        { 5.10480000000000018e+02, 8.82417822255562042e+219 },
        { 5.24660000000000082e+02, 1.25319176676322104e+226 },
        { 5.38839999999999918e+02, 1.78040834846872269e+232 },
        { 5.53019999999999982e+02, 2.53030155402530434e+238 },
        { 5.67200000000000045e+02, 3.59722733532623695e+244 },
        { 5.81379999999999882e+02, 5.11563284594439370e+250 },
        { 5.95559999999999945e+02, 7.27713072718670134e+256 },
        { 6.09740000000000009e+02, 1.03548598244646494e+263 },
        { 6.23920000000000073e+02, 1.47382473286561674e+269 },
        { 6.38099999999999909e+02, 2.09826211076537001e+275 },
        { 6.52279999999999973e+02, 2.98800266537820785e+281 },
        { 6.66460000000000036e+02, 4.25603261442151850e+287 },
        { 6.80639999999999873e+02, 6.06355486411875886e+293 },
        { 6.94819999999999936e+02, 8.64060224208150321e+299 },
        // N(0, 1e-4) grid
        { -2.33376857480713804e-04, 1.00000001361618951e+00 },
        { -2.06191650080946232e-04, 1.00000001062874921e+00 },
        { -1.88950996033343028e-04, 1.00000000892561980e+00 },
        { -1.75986102793353330e-04, 1.00000000774277709e+00 },
        { -1.65443472569025288e-04, 1.00000000684288559e+00 },
        { -1.56472647136179878e-04, 1.00000000612092244e+00 },
        { -1.48609156933045646e-04, 1.00000000552117041e+00 },
        { -1.41570209380050639e-04, 1.00000000501053110e+00 },
        { -1.35170223993066169e-04, 1.00000000456774729e+00 },
        { -1.29280522919941596e-04, 1.00000000417836343e+00 },
        { -1.23808033554710786e-04, 1.00000000383210730e+00 },
        { -1.18683143275581803e-04, 1.00000000352142204e+00 },
        { -1.13852347846006915e-04, 1.00000000324058935e+00 },
        { -1.09273582872572116e-04, 1.00000000298517899e+00 },
        { -1.04913139796397118e-04, 1.00000000275169176e+00 },
        { -1.00743560128730786e-04, 1.00000000253731613e+00 },
        { -9.67421566101701072e-05, 1.00000000233976127e+00 },
        { -9.28899491647270785e-05, 1.00000000215713558e+00 },
        { -8.91708829673407106e-05, 1.00000000198786165e+00 },
        { -8.55712430592972013e-05, 1.00000000183060944e+00 },
        { -8.20792088332380403e-05, 1.00000000168424918e+00 },
        { -7.86845099455004798e-05, 1.00000000154781299e+00 },
        { -7.53781570273144461e-05, 1.00000000142046663e+00 },
        { -7.21522283982343359e-05, 1.00000000130148603e+00 },
        { -6.89996992858782750e-05, 1.00000000119023968e+00 },
        { -6.59143037113206254e-05, 1.00000000108617382e+00 },
        { -6.28904217632189838e-05, 1.00000000098880126e+00 },
        { -5.99229868099344597e-05, 1.00000000089769103e+00 },
        { -5.70074085182460576e-05, 1.00000000081246121e+00 },
        { -5.41395085129087674e-05, 1.00000000073277162e+00 },
        { -5.13154662265951307e-05, 1.00000000065831918e+00 },
        { -4.85317730254962147e-05, 1.00000000058883320e+00 },
        { -4.57851931012495048e-05, 1.00000000052407101e+00 },
        { -4.30727299295457583e-05, 1.00000000046381499e+00 },
        { -4.03915973344409098e-05, 1.00000000040787018e+00 },
        { -3.77391943828553784e-05, 1.00000000035606162e+00 },
        { -3.51130834789464396e-05, 1.00000000030823211e+00 },
        { -3.25109711425282096e-05, 1.00000000026424085e+00 },
        { -2.99306910465667103e-05, 1.00000000022396152e+00 },
        { -2.73701889612983405e-05, 1.00000000018728175e+00 },
        { -2.48275093107116347e-05, 1.00000000015410140e+00 },
        { -2.23007830940366815e-05, 1.00000000012433121e+00 },
        { -1.97882169628358073e-05, 1.00000000009789347e+00 },
        { -1.72880832750766922e-05, 1.00000000007471956e+00 },
        { -1.47987109725838876e-05, 1.00000000005475043e+00 },
        { -1.23184771485879421e-05, 1.00000000003793632e+00 },
        { -9.84579918856374262e-06, 1.00000000002423484e+00 },
        { -7.37912738082726998e-06, 1.00000000001361289e+00 },
        { -4.91693790395577869e-06, 1.00000000000604405e+00 },
        { -2.45772610640276230e-06, 1.00000000000151013e+00 },
        { 0.00000000000000000e+00, 1.00000000000000000e+00 },
        { 2.45772610640274874e-06, 1.00000000000151013e+00 },
        { 4.91693790395579224e-06, 1.00000000000604405e+00 },
        { 7.37912738082726998e-06, 1.00000000001361289e+00 },
        { 9.84579918856372907e-06, 1.00000000002423484e+00 },
        { 1.23184771485879556e-05, 1.00000000003793632e+00 },
        { 1.47987109725838876e-05, 1.00000000005475043e+00 },
        { 1.72880832750766922e-05, 1.00000000007471956e+00 },
        { 1.97882169628358174e-05, 1.00000000009789347e+00 },
        { 2.23007830940366815e-05, 1.00000000012433121e+00 },
        { 2.48275093107116347e-05, 1.00000000015410140e+00 },
        { 2.73701889612983235e-05, 1.00000000018728175e+00 },
        { 2.99306910465667103e-05, 1.00000000022396152e+00 },
        { 3.25109711425282096e-05, 1.00000000026424085e+00 },
        { 3.51130834789464260e-05, 1.00000000030823211e+00 },
        { 3.77391943828553919e-05, 1.00000000035606162e+00 },
        { 4.03915973344409098e-05, 1.00000000040787018e+00 },
        { 4.30727299295457448e-05, 1.00000000046381499e+00 },
        { 4.57851931012495184e-05, 1.00000000052407101e+00 },
        { 4.85317730254962147e-05, 1.00000000058883320e+00 },
        { 5.13154662265951172e-05, 1.00000000065831918e+00 },
        { 5.41395085129087945e-05, 1.00000000073277162e+00 },
        { 5.70074085182460576e-05, 1.00000000081246121e+00 },
        { 5.99229868099344597e-05, 1.00000000089769103e+00 },
        { 6.28904217632189974e-05, 1.00000000098880126e+00 },
        { 6.59143037113206254e-05, 1.00000000108617382e+00 },
        { 6.89996992858782750e-05, 1.00000000119023968e+00 },
        { 7.21522283982342953e-05, 1.00000000130148603e+00 },
        { 7.53781570273144596e-05, 1.00000000142046663e+00 },
        { 7.86845099455004798e-05, 1.00000000154781299e+00 },
        { 8.20792088332380267e-05, 1.00000000168424918e+00 },
        { 8.55712430592972013e-05, 1.00000000183060944e+00 },
        { 8.91708829673407106e-05, 1.00000000198786165e+00 },
        { 9.28899491647270650e-05, 1.00000000215713558e+00 },
        { 9.67421566101701207e-05, 1.00000000233976127e+00 },
        { 1.00743560128730786e-04, 1.00000000253731613e+00 },
        { 1.04913139796397118e-04, 1.00000000275169176e+00 },
        { 1.09273582872572130e-04, 1.00000000298517899e+00 },
        { 1.13852347846006915e-04, 1.00000000324058935e+00 },
        { 1.18683143275581790e-04, 1.00000000352142204e+00 },
        { 1.23808033554710813e-04, 1.00000000383210730e+00 },
        { 1.29280522919941596e-04, 1.00000000417836343e+00 },
        { 1.35170223993066142e-04, 1.00000000456774729e+00 },
        { 1.41570209380050585e-04, 1.00000000501053110e+00 },
        { 1.48609156933045673e-04, 1.00000000552117041e+00 },
        { 1.56472647136179878e-04, 1.00000000612092244e+00 },
        { 1.65443472569025234e-04, 1.00000000684288559e+00 },
        { 1.75986102793353384e-04, 1.00000000774277709e+00 },
        { 1.88950996033343001e-04, 1.00000000892561980e+00 },
        { 2.06191650080946124e-04, 1.00000001062874921e+00 },
        // N(15, 1e-4) grid
        { 1.49997666231425200e+01, 3.39572805187862541e+05 },
        { 1.49997938083499189e+01, 3.39581723433457839e+05 },
        { 1.49998110490039664e+01, 3.39587379439403943e+05 },
        { 1.49998240138972072e+01, 3.39591632792299381e+05 },
        { 1.49998345565274303e+01, 3.39595091520141170e+05 },
        { 1.49998435273528639e+01, 3.39598034612729505e+05 },
        { 1.49998513908430677e+01, 3.39600614438695717e+05 },
        { 1.49998584297906206e+01, 3.39602923768298875e+05 },
        { 1.49998648297760067e+01, 3.39605023481612618e+05 },
        { 1.49998707194770802e+01, 3.39606955791861168e+05 },
        { 1.49998761919664449e+01, 3.39608751231957809e+05 },
        { 1.49998813168567242e+01, 3.39610432638736849e+05 },
        { 1.49998861476521537e+01, 3.39612017564637295e+05 },
        { 1.49998907264171280e+01, 3.39613519809220335e+05 },
        { 1.49998950868602030e+01, 3.39614950430879253e+05 },
        { 1.49998992564398712e+01, 3.39616318437643233e+05 },
        { 1.49999032578433891e+01, 3.39617631272310507e+05 },
        { 1.49999071100508345e+01, 3.39618895161515858e+05 },
        { 1.49999108291170327e+01, 3.39620115372244734e+05 },
        { 1.49999144287569415e+01, 3.39621296403871733e+05 },
        { 1.49999179207911659e+01, 3.39622442134313867e+05 },
        { 1.49999213154900541e+01, 3.39623555932913208e+05 },
        { 1.49999246218429718e+01, 3.39624640748774516e+05 },
        { 1.49999278477716018e+01, 3.39625699180739291e+05 },
        { 1.49999310003007142e+01, 3.39626733533408202e+05 },
        { 1.49999340856962888e+01, 3.39627745862453652e+05 },
        { 1.49999371095782372e+01, 3.39628738011598936e+05 },
        { 1.49999400770131892e+01, 3.39629711643057759e+05 },
        { 1.49999429925914818e+01, 3.39630668262791296e+05 },
        { 1.49999458604914864e+01, 3.39631609241611266e+05 },
        { 1.49999486845337735e+01, 3.39632535832949507e+05 },
        { 1.49999514682269748e+01, 3.39633449187903258e+05 },
        { 1.49999542148068983e+01, 3.39634350368067739e+05 },
        { 1.49999569272700697e+01, 3.39635240356539376e+05 },
        { 1.49999596084026656e+01, 3.39636120067408367e+05 },
        { 1.49999622608056171e+01, 3.39636990353993606e+05 },
        { 1.49999648869165210e+01, 3.39637852016031509e+05 },
        { 1.49999674890288581e+01, 3.39638705805980950e+05 },
        { 1.49999700693089526e+01, 3.39639552434586571e+05 },
        { 1.49999726298110385e+01, 3.39640392575822247e+05 },
        { 1.49999751724906893e+01, 3.39641226871296705e+05 },
        { 1.49999776992169060e+01, 3.39642055934219272e+05 },
        { 1.49999802117830363e+01, 3.39642880352982378e+05 },
        { 1.49999827119167257e+01, 3.39643700694426079e+05 },
        { 1.49999852012890269e+01, 3.39644517506827018e+05 },
        { 1.49999876815228514e+01, 3.39645331322670914e+05 },
        { 1.49999901542008107e+01, 3.39646142661225575e+05 },
        { 1.49999926208726198e+01, 3.39646952030969667e+05 },
        { 1.49999950830620961e+01, 3.39647759931889130e+05 },
        { 1.49999975422738938e+01, 3.39648566857686790e+05 },
        { 1.50000000000000000e+01, 3.39649373297913873e+05 },
        { 1.50000024577261062e+01, 3.39650179740060470e+05 },
        { 1.50000049169379039e+01, 3.39650986671621271e+05 },
        { 1.50000073791273802e+01, 3.39651794582161412e+05 },
        { 1.50000098457991893e+01, 3.39652603965407121e+05 },
        { 1.50000123184771486e+01, 3.39653415321377455e+05 },
        { 1.50000147987109731e+01, 3.39654229158593924e+05 },
        { 1.50000172880832743e+01, 3.39655045996377361e+05 },
        { 1.50000197882169637e+01, 3.39655866367277456e+05 },
        { 1.50000223007830940e+01, 3.39656690819645592e+05 },
        { 1.50000248275093107e+01, 3.39657519920408842e+05 },
        { 1.50000273701889615e+01, 3.39658354258058942e+05 },
        { 1.50000299306910474e+01, 3.39659194445918198e+05 },
        { 1.50000325109711419e+01, 3.39660041125722695e+05 },
        { 1.50000351130834790e+01, 3.39660894971589383e+05 },
        { 1.50000377391943829e+01, 3.39661756694423326e+05 },
        { 1.50000403915973344e+01, 3.39662627046862326e+05 },
        { 1.50000430727299303e+01, 3.39663506828842452e+05 },
        { 1.50000457851931017e+01, 3.39664396893905359e+05 },
        { 1.50000485317730252e+01, 3.39665298156388977e+05 },
        { 1.50000513154662265e+01, 3.39666211599666276e+05 },
        { 1.50000541395085136e+01, 3.39667138285640627e+05 },
        { 1.50000570074085182e+01, 3.39668079365753860e+05 },
        { 1.50000599229868108e+01, 3.39669036093822913e+05 },
        { 1.50000628904217628e+01, 3.39670009841091698e+05 },
        { 1.50000659143037112e+01, 3.39671002114007017e+05 },
        { 1.50000689996992858e+01, 3.39672014575330191e+05 },
        { 1.50000721522283982e+01, 3.39673049069404136e+05 },
        { 1.50000753781570282e+01, 3.39674107652604813e+05 },
        { 1.50000786845099459e+01, 3.39675192630335805e+05 },
        { 1.50000820792088341e+01, 3.39676306602358818e+05 },
        { 1.50000855712430585e+01, 3.39677452518839098e+05 },
        { 1.50000891708829673e+01, 3.39678633750348934e+05 },
        { 1.50000928899491655e+01, 3.39679854176241672e+05 },
        { 1.50000967421566109e+01, 3.39681118297581968e+05 },
        { 1.50001007435601288e+01, 3.39682431383360876e+05 },
        { 1.50001049131397970e+01, 3.39683799662617035e+05 },
        { 1.50001092735828720e+01, 3.39685230581061041e+05 },
        { 1.50001138523478463e+01, 3.39686733150295448e+05 },
        { 1.50001186831432758e+01, 3.39688318433161825e+05 },
        { 1.50001238080335551e+01, 3.39690000234851963e+05 },
        { 1.50001292805229198e+01, 3.39691796115073957e+05 },
        { 1.50001351702239933e+01, 3.39693728920267487e+05 },
        { 1.50001415702093794e+01, 3.39695829196402454e+05 },
        { 1.50001486091569323e+01, 3.39698139175078017e+05 },
        { 1.50001564726471361e+01, 3.39700719763387402e+05 },
        { 1.50001654434725697e+01, 3.39703663773661538e+05 },
        { 1.50001759861027928e+01, 3.39707123645353131e+05 },
        { 1.50001889509960336e+01, 3.39711378501756873e+05 },
        { 1.50002061916500811e+01, 3.39717036672546528e+05 },
    };
    static const double i1e[][2] = {
        // Uniform [-709, +709] grid
        { -7.09000000000000000e+02, -1.23067888965247857e+306 },
        { -6.94820000000000050e+02, -8.63438213064202749e+299 },
        { -6.80639999999999986e+02, -6.05909892170071920e+293 },
        { -6.66460000000000036e+02, -4.25283840057319577e+287 },
        { -6.52279999999999973e+02, -2.98571135670705352e+281 },
        { -6.38100000000000023e+02, -2.09661731725642791e+275 },
        { -6.23919999999999959e+02, -1.47264315811858754e+269 },
        { -6.09740000000000009e+02, -1.03463651280285006e+263 },
        { -5.95559999999999945e+02, -7.27101867247115274e+256 },
        { -5.81379999999999995e+02, -5.11123139042528858e+250 },
        { -5.67200000000000045e+02, -3.59405489557111873e+244 },
        { -5.53019999999999982e+02, -2.52801280544068546e+238 },
        { -5.38840000000000032e+02, -1.77875550539537887e+232 },
        { -5.24659999999999968e+02, -1.25199690718446768e+226 },
        { -5.10480000000000018e+02, -8.81553096093478104e+219 },
        { -4.96300000000000011e+02, -6.20955983760586136e+213 },
        { -4.82120000000000005e+02, -4.37572434810839614e+207 },
        { -4.67939999999999998e+02, -3.08479588359686804e+201 },
        { -4.53759999999999991e+02, -2.17571376937298307e+195 },
        { -4.39579999999999984e+02, -1.53528336837751007e+189 },
        { -4.25399999999999977e+02, -1.08392849409101527e+183 },
        { -4.11220000000000027e+02, -7.65690528636694960e+176 },
        { -3.97040000000000020e+02, -5.41206900666695048e+170 },
        { -3.82860000000000014e+02, -3.82780208313026895e+164 },
        { -3.68680000000000007e+02, -2.70914671287504912e+158 },
        { -3.54500000000000000e+02, -1.91882656490666564e+152 },
        { -3.40319999999999993e+02, -1.36014483002938342e+146 },
        { -3.26139999999999986e+02, -9.64962053360065347e+139 },
        { -3.11960000000000036e+02, -6.85242498535870676e+133 },
        { -2.97780000000000030e+02, -4.87108028740881905e+127 },
        { -2.83600000000000023e+02, -3.46654421341918490e+121 },
        { -2.69420000000000016e+02, -2.47006768219344092e+115 },
        { -2.55240000000000009e+02, -1.76246270903799677e+109 },
        { -2.41060000000000002e+02, -1.25950034169699536e+103 },
        { -2.26879999999999995e+02, -9.01622204982702926e+96 },
        { -2.12699999999999989e+02, -6.46688492478025839e+90 },
        { -1.98519999999999982e+02, -4.64864125063932631e+84 },
        { -1.84340000000000032e+02, -3.35011088529169346e+78 },
        { -1.70159999999999968e+02, -2.42142175167506762e+72 },
        { -1.55980000000000018e+02, -1.75623010804411146e+66 },
        { -1.41799999999999955e+02, -1.27901867944677480e+60 },
        { -1.27620000000000005e+02, -9.36119530535991435e+53 },
        { -1.13440000000000055e+02, -6.89367737060279837e+47 },
        { -9.92599999999999909e+01, -5.11616001962349727e+41 },
        { -8.50800000000000409e+01, -3.83570567567467213e+35 },
        { -7.08999999999999773e+01, -2.91576331942750496e+29 },
        { -5.67200000000000273e+01, -2.26114260037075141e+23 },
        { -4.25399999999999636e+01, -1.80935682856467456e+17 },
        { -2.83600000000000136e+01, -1.53215015386715179e+11 },
        { -1.41800000000000637e+01, -1.48405470354675519e+05 },
        { 0.00000000000000000e+00, 0.00000000000000000e+00 },
        { 1.41799999999999500e+01, 1.48405470354659221e+05 },
        { 2.83600000000000136e+01, 1.53215015386715179e+11 },
        { 4.25399999999999636e+01, 1.80935682856467456e+17 },
        { 5.67200000000000273e+01, 2.26114260037075141e+23 },
        { 7.08999999999999773e+01, 2.91576331942750496e+29 },
        { 8.50799999999999272e+01, 3.83570567567423900e+35 },
        { 9.92599999999999909e+01, 5.11616001962349727e+41 },
        { 1.13439999999999941e+02, 6.89367737060201790e+47 },
        { 1.27620000000000005e+02, 9.36119530535991435e+53 },
        { 1.41799999999999955e+02, 1.27901867944677480e+60 },
        { 1.55980000000000018e+02, 1.75623010804411146e+66 },
        { 1.70159999999999968e+02, 2.42142175167506762e+72 },
        { 1.84340000000000032e+02, 3.35011088529169346e+78 },
        { 1.98519999999999982e+02, 4.64864125063932631e+84 },
        { 2.12699999999999932e+02, 6.46688492477989202e+90 },
        { 2.26879999999999995e+02, 9.01622204982702926e+96 },
        { 2.41059999999999945e+02, 1.25950034169692395e+103 },
        { 2.55240000000000009e+02, 1.76246270903799677e+109 },
        { 2.69419999999999959e+02, 2.47006768219330093e+115 },
        { 2.83600000000000023e+02, 3.46654421341918490e+121 },
        { 2.97779999999999973e+02, 4.87108028740854201e+127 },
        { 3.11960000000000036e+02, 6.85242498535870676e+133 },
        { 3.26139999999999873e+02, 9.64962053359955770e+139 },
        { 3.40319999999999936e+02, 1.36014483002930623e+146 },
        { 3.54500000000000000e+02, 1.91882656490666564e+152 },
        { 3.68680000000000064e+02, 2.70914671287520277e+158 },
        { 3.82859999999999900e+02, 3.82780208312983420e+164 },
        { 3.97039999999999964e+02, 5.41206900666664371e+170 },
        { 4.11220000000000027e+02, 7.65690528636694960e+176 },
        { 4.25400000000000091e+02, 1.08392849409113840e+183 },
        { 4.39579999999999927e+02, 1.53528336837742301e+189 },
        { 4.53759999999999991e+02, 2.17571376937298307e+195 },
        { 4.67940000000000055e+02, 3.08479588359704317e+201 },
        { 4.82119999999999891e+02, 4.37572434810789940e+207 },
        { 4.96299999999999955e+02, 6.20955983760550973e+213 },
        { 5.10480000000000018e+02, 8.81553096093478104e+219 },
        { 5.24660000000000082e+02, 1.25199690718460997e+226 },
        { 5.38839999999999918e+02, 1.77875550539517672e+232 },
        { 5.53019999999999982e+02, 2.52801280544068546e+238 },
        { 5.67200000000000045e+02, 3.59405489557111873e+244 },
        { 5.81379999999999882e+02, 5.11123139042470799e+250 },
        { 5.95559999999999945e+02, 7.27101867247115274e+256 },
        { 6.09740000000000009e+02, 1.03463651280285006e+263 },
        { 6.23920000000000073e+02, 1.47264315811875470e+269 },
        { 6.38099999999999909e+02, 2.09661731725618990e+275 },
        { 6.52279999999999973e+02, 2.98571135670705352e+281 },
        { 6.66460000000000036e+02, 4.25283840057319577e+287 },
        { 6.80639999999999873e+02, 6.05909892170003070e+293 },
        { 6.94819999999999936e+02, 8.63438213064104606e+299 },
        // N(0, 1e-4) grid
        { -2.33376857480713804e-04, -1.16688429534782774e-04 },
        { -2.06191650080946232e-04, -1.03095825588362948e-04 },
        { -1.88950996033343028e-04, -9.44754984382976948e-05 },
        { -1.75986102793353330e-04, -8.79930517373319574e-05 },
        { -1.65443472569025288e-04, -8.27217365675403415e-05 },
        { -1.56472647136179878e-04, -7.82363238075291717e-05 },
        { -1.48609156933045646e-04, -7.43045786716469488e-05 },
        { -1.41570209380050639e-04, -7.07851048673608014e-05 },
        { -1.35170223993066169e-04, -6.75851121508889421e-05 },
        { -1.29280522919941596e-04, -6.46402615950160528e-05 },
        { -1.23808033554710786e-04, -6.19040168959668124e-05 },
        { -1.18683143275581803e-04, -5.93415717422742656e-05 },
        { -1.13852347846006915e-04, -5.69261740152406294e-05 },
        { -1.09273582872572116e-04, -5.46367915178363575e-05 },
        { -1.04913139796397118e-04, -5.24565699703707127e-05 },
        { -1.00743560128730786e-04, -5.03717801282699631e-05 },
        { -9.67421566101701072e-05, -4.83710783616734410e-05 },
        { -9.28899491647270785e-05, -4.64449746324575963e-05 },
        { -8.91708829673407106e-05, -4.45854415279852014e-05 },
        { -8.55712430592972013e-05, -4.27856215688104846e-05 },
        { -8.20792088332380403e-05, -4.10396044511794823e-05 },
        { -7.86845099455004798e-05, -3.93422550031974694e-05 },
        { -7.53781570273144461e-05, -3.76890785404252599e-05 },
        { -7.21522283982343359e-05, -3.60761142225934450e-05 },
        { -6.89996992858782750e-05, -3.44998496634706808e-05 },
        { -6.59143037113206254e-05, -3.29571518735589117e-05 },
        { -6.28904217632189838e-05, -3.14452108971560227e-05 },
        { -5.99229868099344597e-05, -2.99614934184153111e-05 },
        { -5.70074085182460576e-05, -2.85037042707021047e-05 },
        { -5.41395085129087674e-05, -2.70697542663723569e-05 },
        { -5.13154662265951307e-05, -2.56577331217430565e-05 },
        { -4.85317730254962147e-05, -2.42658865198923864e-05 },
        { -4.57851931012495048e-05, -2.28925965566234262e-05 },
        { -4.30727299295457583e-05, -2.15363649697673242e-05 },
        { -4.03915973344409098e-05, -2.01957986713390882e-05 },
        { -3.77391943828553784e-05, -1.88695971947870591e-05 },
        { -3.51130834789464396e-05, -1.75565417421789649e-05 },
        { -3.25109711425282096e-05, -1.62554855734117873e-05 },
        { -2.99306910465667103e-05, -1.49653455249591861e-05 },
        { -2.73701889612983405e-05, -1.36850944819306548e-05 },
        { -2.48275093107116347e-05, -1.24137546563123056e-05 },
        { -2.23007830940366815e-05, -1.11503915477115118e-05 },
        { -1.97882169628358073e-05, -9.89410848190218794e-06 },
        { -1.72880832750766922e-05, -8.64404163786128587e-06 },
        { -1.47987109725838876e-05, -7.39935548649450327e-06 },
        { -1.23184771485879421e-05, -6.15923857441079975e-06 },
        { -9.84579918856374262e-06, -4.92289959434152445e-06 },
        { -7.37912738082726998e-06, -3.68956369043874782e-06 },
        { -4.91693790395577869e-06, -2.45846895198531909e-06 },
        { -2.45772610640276230e-06, -1.22886305320230907e-06 },
        { 0.00000000000000000e+00, 0.00000000000000000e+00 },
        { 2.45772610640274874e-06, 1.22886305320230230e-06 },
        { 4.91693790395579224e-06, 2.45846895198532587e-06 },
        { 7.37912738082726998e-06, 3.68956369043874782e-06 },
        { 9.84579918856372907e-06, 4.92289959434151768e-06 },
        { 1.23184771485879556e-05, 6.15923857441080653e-06 },
        { 1.47987109725838876e-05, 7.39935548649450327e-06 },
        { 1.72880832750766922e-05, 8.64404163786128587e-06 },
        { 1.97882169628358174e-05, 9.89410848190219302e-06 },
        { 2.23007830940366815e-05, 1.11503915477115118e-05 },
        { 2.48275093107116347e-05, 1.24137546563123056e-05 },
        { 2.73701889612983235e-05, 1.36850944819306464e-05 },
        { 2.99306910465667103e-05, 1.49653455249591861e-05 },
        { 3.25109711425282096e-05, 1.62554855734117873e-05 },
        { 3.51130834789464260e-05, 1.75565417421789581e-05 },
        { 3.77391943828553919e-05, 1.88695971947870659e-05 },
        { 4.03915973344409098e-05, 2.01957986713390882e-05 },
        { 4.30727299295457448e-05, 2.15363649697673175e-05 },
        { 4.57851931012495184e-05, 2.28925965566234330e-05 },
        { 4.85317730254962147e-05, 2.42658865198923864e-05 },
        { 5.13154662265951172e-05, 2.56577331217430497e-05 },
        { 5.41395085129087945e-05, 2.70697542663723704e-05 },
        { 5.70074085182460576e-05, 2.85037042707021047e-05 },
        { 5.99229868099344597e-05, 2.99614934184153111e-05 },
        { 6.28904217632189974e-05, 3.14452108971560295e-05 },
        { 6.59143037113206254e-05, 3.29571518735589117e-05 },
        { 6.89996992858782750e-05, 3.44998496634706808e-05 },
        { 7.21522283982342953e-05, 3.60761142225934247e-05 },
        { 7.53781570273144596e-05, 3.76890785404252666e-05 },
        { 7.86845099455004798e-05, 3.93422550031974694e-05 },
        { 8.20792088332380267e-05, 4.10396044511794755e-05 },
        { 8.55712430592972013e-05, 4.27856215688104846e-05 },
        { 8.91708829673407106e-05, 4.45854415279852014e-05 },
        { 9.28899491647270650e-05, 4.64449746324575895e-05 },
        { 9.67421566101701207e-05, 4.83710783616734478e-05 },
        { 1.00743560128730786e-04, 5.03717801282699631e-05 },
        { 1.04913139796397118e-04, 5.24565699703707127e-05 },
        { 1.09273582872572130e-04, 5.46367915178363643e-05 },
        { 1.13852347846006915e-04, 5.69261740152406294e-05 },
        { 1.18683143275581790e-04, 5.93415717422742588e-05 },
        { 1.23808033554710813e-04, 6.19040168959668260e-05 },
        { 1.29280522919941596e-04, 6.46402615950160528e-05 },
        { 1.35170223993066142e-04, 6.75851121508889286e-05 },
        { 1.41570209380050585e-04, 7.07851048673607743e-05 },
        { 1.48609156933045673e-04, 7.43045786716469624e-05 },
        { 1.56472647136179878e-04, 7.82363238075291717e-05 },
        { 1.65443472569025234e-04, 8.27217365675403144e-05 },
        { 1.75986102793353384e-04, 8.79930517373319845e-05 },
        { 1.88950996033343001e-04, 9.44754984382976812e-05 },
        { 2.06191650080946124e-04, 1.03095825588362893e-04 },
        // N(15, 1e-4) grid
        { 1.49997666231425200e+01, 3.28050769182148448e+05 },
        { 1.49997938083499189e+01, 3.28059406102088222e+05 },
        { 1.49998110490039664e+01, 3.28064883689829614e+05 },
        { 1.49998240138972072e+01, 3.28069002871212957e+05 },
        { 1.49998345565274303e+01, 3.28072352493975952e+05 },
        { 1.49998435273528639e+01, 3.28075202747183037e+05 },
        { 1.49998513908430677e+01, 3.28077701192997920e+05 },
        { 1.49998584297906206e+01, 3.28079937675243418e+05 },
        { 1.49998648297760067e+01, 3.28081971153532795e+05 },
        { 1.49998707194770802e+01, 3.28083842509475711e+05 },
        { 1.49998761919664449e+01, 3.28085581312826020e+05 },
        { 1.49998813168567242e+01, 3.28087209680031170e+05 },
        { 1.49998861476521537e+01, 3.28088744609839458e+05 },
        { 1.49998907264171280e+01, 3.28090199466505786e+05 },
        { 1.49998950868602030e+01, 3.28091584959587315e+05 },
        { 1.49998992564398712e+01, 3.28092909812955535e+05 },
        { 1.49999032578433891e+01, 3.28094181234626099e+05 },
        { 1.49999071100508345e+01, 3.28095405254815356e+05 },
        { 1.49999108291170327e+01, 3.28096586974361970e+05 },
        { 1.49999144287569415e+01, 3.28097730750708317e+05 },
        { 1.49999179207911659e+01, 3.28098840339442890e+05 },
        { 1.49999213154900541e+01, 3.28099919003622257e+05 },
        { 1.49999246218429718e+01, 3.28100969599321892e+05 },
        { 1.49999278477716018e+01, 3.28101994643403159e+05 },
        { 1.49999310003007142e+01, 3.28102996367768501e+05 },
        { 1.49999340856962888e+01, 3.28103976763244369e+05 },
        { 1.49999371095782372e+01, 3.28104937615393952e+05 },
        { 1.49999400770131892e+01, 3.28105880533996795e+05 },
        { 1.49999429925914818e+01, 3.28106806977508648e+05 },
        { 1.49999458604914864e+01, 3.28107718273499282e+05 },
        { 1.49999486845337735e+01, 3.28108615635861177e+05 },
        { 1.49999514682269748e+01, 3.28109500179380411e+05 },
        { 1.49999542148068983e+01, 3.28110372932164406e+05 },
        { 1.49999569272700697e+01, 3.28111234846298001e+05 },
        { 1.49999596084026656e+01, 3.28112086807036540e+05 },
        { 1.49999622608056171e+01, 3.28112929640781251e+05 },
        { 1.49999648869165210e+01, 3.28113764122041001e+05 },
        { 1.49999674890288581e+01, 3.28114590979538450e+05 },
        { 1.49999700693089526e+01, 3.28115410901598108e+05 },
        { 1.49999726298110385e+01, 3.28116224540933501e+05 },
        { 1.49999751724906893e+01, 3.28117032518913911e+05 },
        { 1.49999776992169060e+01, 3.28117835429405153e+05 },
        { 1.49999802117830363e+01, 3.28118633842239040e+05 },
        { 1.49999827119167257e+01, 3.28119428306374932e+05 },
        { 1.49999852012890269e+01, 3.28120219352794054e+05 },
        { 1.49999876815228514e+01, 3.28121007497185259e+05 },
        { 1.49999901542008107e+01, 3.28121793242436193e+05 },
        { 1.49999926208726198e+01, 3.28122577080985648e+05 },
        { 1.49999950830620961e+01, 3.28123359497047786e+05 },
        { 1.49999975422738938e+01, 3.28124140968751919e+05 },
        { 1.50000000000000000e+01, 3.28124921970206371e+05 },
        { 1.50000024577261062e+01, 3.28125702973523759e+05 },
        { 1.50000049169379039e+01, 3.28126484450820892e+05 },
        { 1.50000073791273802e+01, 3.28127266876219946e+05 },
        { 1.50000098457991893e+01, 3.28128050727872702e+05 },
        { 1.50000123184771486e+01, 3.28128836490025395e+05 },
        { 1.50000147987109731e+01, 3.28129624655158550e+05 },
        { 1.50000172880832743e+01, 3.28130415726211271e+05 },
        { 1.50000197882169637e+01, 3.28131210218934342e+05 },
        { 1.50000223007830940e+01, 3.28132008664381749e+05 },
        { 1.50000248275093107e+01, 3.28132811611597193e+05 },
        { 1.50000273701889615e+01, 3.28133619630508882e+05 },
        { 1.50000299306910474e+01, 3.28134433315092116e+05 },
        { 1.50000325109711419e+01, 3.28135253286839987e+05 },
        { 1.50000351130834790e+01, 3.28136080198604846e+05 },
        { 1.50000377391943829e+01, 3.28136914738866850e+05 },
        { 1.50000403915973344e+01, 3.28137757636522234e+05 },
        { 1.50000430727299303e+01, 3.28138609666273813e+05 },
        { 1.50000457851931017e+01, 3.28139471654738765e+05 },
        { 1.50000485317730252e+01, 3.28140344487413066e+05 },
        { 1.50000513154662265e+01, 3.28141229116649774e+05 },
        { 1.50000541395085136e+01, 3.28142126570855558e+05 },
        { 1.50000570074085182e+01, 3.28143037965150666e+05 },
        { 1.50000599229868108e+01, 3.28143964513801620e+05 },
        { 1.50000628904217628e+01, 3.28144907544797345e+05 },
        { 1.50000659143037112e+01, 3.28145868517065130e+05 },
        { 1.50000689996992858e+01, 3.28146849040915782e+05 },
        { 1.50000721522283982e+01, 3.28147850902513950e+05 },
        { 1.50000753781570282e+01, 3.28148876093368861e+05 },
        { 1.50000786845099459e+01, 3.28149926846162183e+05 },
        { 1.50000820792088341e+01, 3.28151005678648187e+05 },
        { 1.50000855712430585e+01, 3.28152115447931748e+05 },
        { 1.50000891708829673e+01, 3.28153259418263333e+05 },
        { 1.50000928899491655e+01, 3.28154441346625332e+05 },
        { 1.50000967421566109e+01, 3.28155665592100180e+05 },
        { 1.50001007435601288e+01, 3.28156937257473299e+05 },
        { 1.50001049131397970e+01, 3.28158262375293649e+05 },
        { 1.50001092735828720e+01, 3.28159648156403448e+05 },
        { 1.50001138523478463e+01, 3.28161103328142199e+05 },
        { 1.50001186831432758e+01, 3.28162638604383974e+05 },
        { 1.50001238080335551e+01, 3.28164267354848213e+05 },
        { 1.50001292805229198e+01, 3.28166006585337687e+05 },
        { 1.50001351702239933e+01, 3.28167878421622037e+05 },
        { 1.50001415702093794e+01, 3.28169912446126342e+05 },
        { 1.50001486091569323e+01, 3.28172149558293051e+05 },
        { 1.50001564726471361e+01, 3.28174648743957805e+05 },
        { 1.50001654434725697e+01, 3.28177499887774000e+05 },
        { 1.50001759861027928e+01, 3.28180850620636949e+05 },
        { 1.50001889509960336e+01, 3.28184971261166851e+05 },
        { 1.50002061916500811e+01, 3.28190450949877151e+05 },
    };
    const double epsilon = std::numeric_limits<double>::epsilon();
    size_t i;

    using numerics::i_0;
    using numerics::i_1;

    for (i = 0; i < arraycount(i0e); i++) {
        const double x = i0e[i][0];
        const double y = i0e[i][1];
        assert(y == 0 ? i_0(x) == 0 : relerr(i_0(x), y) < 10*epsilon);
    }

    for (i = 0; i < arraycount(i1e); i++) {
        const double x = i1e[i][0];
        const double y = i1e[i][1];
        assert(y == 0 ? i_1(x) == 0 : relerr(i_1(x), y) < 10*epsilon);
    }
}

int main(int argc, char** argv) {
    test_draw();
    test_linspace();
    test_log_linspace();
    test_logaddexp();
    test_bessel();

    return 0;
}
